@@ -24,9 +24,7 @@
 #include "render.h"
 
 typedef enum {
-#if defined (HAVE_CAIRO)
 	PGD_RENDER_CAIRO,
-#endif
 	PGD_RENDER_PIXBUF
 } PgdRenderMode;
 
@@ -39,6 +37,7 @@ typedef struct {
 	gdouble          scale;
 	gint             rotate;
 	GdkRectangle     slice;
+	gboolean         printing;
 	
 	GtkWidget       *swindow;
 	GtkWidget       *darea;
@@ -47,10 +46,8 @@ typedef struct {
 	GtkWidget       *slice_w;
 	GtkWidget       *slice_h;
 	GtkWidget       *timer_label;
-	
-#if defined (HAVE_CAIRO)
+
 	cairo_surface_t *surface;
-#endif
 	GdkPixbuf       *pixbuf;
 } PgdRenderDemo;
 
@@ -65,12 +62,10 @@ pgd_render_free (PgdRenderDemo *demo)
 		demo->doc = NULL;
 	}
 	
-#if defined (HAVE_CAIRO)
 	if (demo->surface) {
 		cairo_surface_destroy (demo->surface);
 		demo->surface = NULL;
 	}
-#endif
 
 	if (demo->pixbuf) {
 		g_object_unref (demo->pixbuf);
@@ -85,28 +80,24 @@ pgd_render_drawing_area_expose (GtkWidget      *area,
 				GdkEventExpose *event,
 				PgdRenderDemo  *demo)
 {
-#if defined (HAVE_CAIRO)
 	if (demo->mode == PGD_RENDER_CAIRO && !demo->surface)
 		return FALSE;
-#endif
-	
+
 	if (demo->mode == PGD_RENDER_PIXBUF && !demo->pixbuf)
 		return FALSE;
 
-	gdk_window_clear (area->window);
+	gdk_window_clear (gtk_widget_get_window (area));
 
-#if defined (HAVE_CAIRO)
 	if (demo->mode == PGD_RENDER_CAIRO) {
 		cairo_t *cr;
 
-		cr = gdk_cairo_create (area->window);
+		cr = gdk_cairo_create (gtk_widget_get_window (area));
 		cairo_set_source_surface (cr, demo->surface, 0, 0);
 		cairo_paint (cr);
 		cairo_destroy (cr);
 	} else if (demo->mode == PGD_RENDER_PIXBUF) {
-#endif
-		gdk_draw_pixbuf (area->window,
-				 area->style->fg_gc[GTK_STATE_NORMAL],
+		gdk_draw_pixbuf (gtk_widget_get_window (area),
+				 gtk_widget_get_style(area)->fg_gc[GTK_STATE_NORMAL],
 				 demo->pixbuf,
 				 0, 0,
 				 0, 0,
@@ -114,12 +105,10 @@ pgd_render_drawing_area_expose (GtkWidget      *area,
 				 gdk_pixbuf_get_height (demo->pixbuf),
 				 GDK_RGB_DITHER_NORMAL,
 				 0, 0);
-#if defined (HAVE_CAIRO)
 	} else {
 		g_assert_not_reached ();
 	}
-#endif
-	
+
 	return TRUE;
 }
 
@@ -138,12 +127,10 @@ pgd_render_start (GtkButton     *button,
 	if (!page)
 		return;
 
-#if defined (HAVE_CAIRO)
 	if (demo->surface)
 		cairo_surface_destroy (demo->surface);
 	demo->surface = NULL;
-#endif
-	
+
 	if (demo->pixbuf)
 		g_object_unref (demo->pixbuf);
 	demo->pixbuf = NULL;
@@ -162,7 +149,6 @@ pgd_render_start (GtkButton     *button,
 		y = demo->slice.x * demo->scale;
 	}
 
-#if defined (HAVE_CAIRO)
 	if (demo->mode == PGD_RENDER_CAIRO) {
 		cairo_t *cr;
 
@@ -191,8 +177,11 @@ pgd_render_start (GtkButton     *button,
 		
 		if (demo->rotate != 0)
 			cairo_rotate (cr, demo->rotate * G_PI / 180.0);
-		
-		poppler_page_render (page, cr);
+
+		if (demo->printing)
+			poppler_page_render_for_printing (page, cr);
+		else
+			poppler_page_render (page, cr);
 		cairo_restore (cr);
 
 		cairo_set_operator (cr, CAIRO_OPERATOR_DEST_OVER);
@@ -203,26 +192,34 @@ pgd_render_start (GtkButton     *button,
 		
 		cairo_destroy (cr);
 	} else if (demo->mode == PGD_RENDER_PIXBUF) {
-#endif
 #ifdef POPPLER_WITH_GDK
 		timer = g_timer_new ();
 		demo->pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB,
 					       FALSE, 8, width, height);
 		gdk_pixbuf_fill (demo->pixbuf, 0xffffff);
-		poppler_page_render_to_pixbuf (page,
-					       x, y,
-					       width,
-					       height,
-					       demo->scale,
-					       demo->rotate,
-					       demo->pixbuf);
+		if (demo->printing) {
+			poppler_page_render_to_pixbuf_for_printing (page,
+								    x, y,
+								    width,
+								    height,
+								    demo->scale,
+								    demo->rotate,
+								    demo->pixbuf);
+		} else {
+			poppler_page_render_to_pixbuf (page,
+						       x, y,
+						       width,
+						       height,
+						       demo->scale,
+						       demo->rotate,
+						       demo->pixbuf);
+		}
 		g_timer_stop (timer);
 #endif /* POPPLER_WITH_GDK */
-#if defined (HAVE_CAIRO)
 	} else {
 		g_assert_not_reached ();
 	}
-#endif
+
 	g_object_unref (page);
 	
 	str = g_strdup_printf ("<i>Page rendered in %.4f seconds</i>",
@@ -283,6 +280,13 @@ pgd_render_rotate_selector_changed (GtkComboBox   *combobox,
 	demo->rotate = gtk_combo_box_get_active (combobox) * 90;
 }
 
+static void
+pgd_render_printing_selector_changed (GtkToggleButton *tooglebutton,
+				      PgdRenderDemo *demo)
+{
+	demo->printing = gtk_toggle_button_get_active (tooglebutton);
+}
+
 static void
 pgd_render_mode_selector_changed (GtkComboBox   *combobox,
 				  PgdRenderDemo *demo)
@@ -309,6 +313,7 @@ pgd_render_properties_selector_create (PgdRenderDemo *demo)
 	GtkWidget *scale_hbox, *scale_selector;
 	GtkWidget *rotate_hbox, *rotate_selector;
 	GtkWidget *mode_hbox, *mode_selector;
+	GtkWidget *printing_selector;
 	GtkWidget *slice_hbox, *slice_selector;
 	GtkWidget *button;
 	gint       n_pages;
@@ -367,11 +372,19 @@ pgd_render_properties_selector_create (PgdRenderDemo *demo)
 	gtk_box_pack_start (GTK_BOX (rotate_hbox), label, TRUE, TRUE, 0);
 	gtk_widget_show (label);
 
+#if GTK_CHECK_VERSION (2, 24, 0)
+	rotate_selector = gtk_combo_box_text_new ();
+	gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (rotate_selector), "0");
+	gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (rotate_selector), "90");
+	gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (rotate_selector), "180");
+	gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (rotate_selector), "270");
+#else
 	rotate_selector = gtk_combo_box_new_text ();
 	gtk_combo_box_append_text (GTK_COMBO_BOX (rotate_selector), "0");
 	gtk_combo_box_append_text (GTK_COMBO_BOX (rotate_selector), "90");
 	gtk_combo_box_append_text (GTK_COMBO_BOX (rotate_selector), "180");
 	gtk_combo_box_append_text (GTK_COMBO_BOX (rotate_selector), "270");
+#endif
 	gtk_combo_box_set_active (GTK_COMBO_BOX (rotate_selector), 0);
 	g_signal_connect (G_OBJECT (rotate_selector), "changed",
 			  G_CALLBACK (pgd_render_rotate_selector_changed),
@@ -388,13 +401,19 @@ pgd_render_properties_selector_create (PgdRenderDemo *demo)
 	gtk_box_pack_start (GTK_BOX (mode_hbox), label, TRUE, TRUE, 0);
 	gtk_widget_show (label);
 
-	mode_selector = gtk_combo_box_new_text ();
-#if defined (HAVE_CAIRO)
-	gtk_combo_box_append_text (GTK_COMBO_BOX (mode_selector), "cairo");
+#if GTK_CHECK_VERSION (2, 24, 0)
+	mode_selector = gtk_combo_box_text_new ();
+	gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (mode_selector), "cairo");
+#ifdef POPPLER_WITH_GDK
+	gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (mode_selector), "pixbuf");
 #endif
+#else // ! GTK_CHECK_VERSION (2, 24, 0)
+	mode_selector = gtk_combo_box_new_text ();
+	gtk_combo_box_append_text (GTK_COMBO_BOX (mode_selector), "cairo");
 #ifdef POPPLER_WITH_GDK
 	gtk_combo_box_append_text (GTK_COMBO_BOX (mode_selector), "pixbuf");
 #endif
+#endif // GTK_CHECK_VERSION (2, 24, 0)
 	gtk_combo_box_set_active (GTK_COMBO_BOX (mode_selector), 0);
 	g_signal_connect (G_OBJECT (mode_selector), "changed",
 			  G_CALLBACK (pgd_render_mode_selector_changed),
@@ -405,6 +424,13 @@ pgd_render_properties_selector_create (PgdRenderDemo *demo)
 	gtk_box_pack_start (GTK_BOX (hbox), mode_hbox, FALSE, TRUE, 0);
 	gtk_widget_show (mode_hbox);
 
+	printing_selector = gtk_check_button_new_with_label ("Printing");
+	g_signal_connect (printing_selector, "toggled",
+			  G_CALLBACK (pgd_render_printing_selector_changed),
+			  (gpointer)demo);
+	gtk_box_pack_start (GTK_BOX (hbox), printing_selector, FALSE, TRUE, 0);
+	gtk_widget_show (printing_selector);
+
 	hbox = gtk_hbox_new (FALSE, 12);
 	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
 	gtk_widget_show (hbox);
