@@ -14,7 +14,7 @@
 // under GPL version 2 or later
 //
 // Copyright (C) 2005 Kristian Høgsberg <krh@redhat.com>
-// Copyright (C) 2005-2011 Albert Astals Cid <aacid@kde.org>
+// Copyright (C) 2005-2012 Albert Astals Cid <aacid@kde.org>
 // Copyright (C) 2005 Jeff Muizelaar <jrmuizel@nit.ca>
 // Copyright (C) 2005 Jonathan Blandford <jrb@redhat.com>
 // Copyright (C) 2005 Marco Pesenti Gritti <mpg@redhat.com>
@@ -40,6 +40,7 @@
 #include <stdlib.h>
 #include "goo/gmem.h"
 #include "Object.h"
+#include "PDFDoc.h"
 #include "XRef.h"
 #include "Array.h"
 #include "Dict.h"
@@ -57,13 +58,14 @@
 // Catalog
 //------------------------------------------------------------------------
 
-Catalog::Catalog(XRef *xrefA) {
+Catalog::Catalog(PDFDoc *docA) {
   Object catDict, pagesDict, pagesDictRef;
   Object obj, obj2;
   Object optContentProps;
 
   ok = gTrue;
-  xref = xrefA;
+  doc = docA;
+  xref = doc->getXRef();
   pages = NULL;
   pageRefs = NULL;
   numPages = -1;
@@ -87,7 +89,7 @@ Catalog::Catalog(XRef *xrefA) {
 
   xref->getCatalog(&catDict);
   if (!catDict.isDict()) {
-    error(-1, "Catalog object is wrong type (%s)", catDict.getTypeName());
+    error(errSyntaxError, -1, "Catalog object is wrong type ({0:s})", catDict.getTypeName());
     goto err1;
   }
   // get the AcroForm dictionary
@@ -114,11 +116,6 @@ Catalog::Catalog(XRef *xrefA) {
 
   // get the ViewerPreferences dictionary
   catDict.dictLookup("ViewerPreferences", &viewerPreferences);
-
-  // perform form-related loading after all widgets have been loaded
-  if (getForm())
-    getForm()->postWidgetsLoad(this);
-
   catDict.free();
   return;
 
@@ -186,7 +183,7 @@ GooString *Catalog::readMetadata() {
     if (catDict.isDict()) {
       catDict.dictLookup("Metadata", &metadata);
     } else {
-      error(-1, "Catalog object is wrong type (%s)", catDict.getTypeName());
+      error(errSyntaxError, -1, "Catalog object is wrong type ({0:s})", catDict.getTypeName());
       metadata.initNull();
     }
     catDict.free();
@@ -197,7 +194,7 @@ GooString *Catalog::readMetadata() {
   }
   dict = metadata.streamGetDict();
   if (!dict->lookup("Subtype", &obj)->isName("XML")) {
-    error(-1, "Unknown Metadata type: '%s'",
+    error(errSyntaxWarning, -1, "Unknown Metadata type: '{0:s}'",
 	  obj.isName() ? obj.getName() : "???");
   }
   obj.free();
@@ -246,13 +243,13 @@ GBool Catalog::cachePageTree(int page)
         pagesRef = pagesDictRef.getRef();
         pagesDictRef.free();
       } else {
-        error(-1, "Catalog dictionary does not contain a valid \"Pages\" entry");
+        error(errSyntaxError, -1, "Catalog dictionary does not contain a valid \"Pages\" entry");
         pagesDictRef.free();
         catDict.free();
         return gFalse;
       }
     } else {
-      error(-1, "Could not find catalog dictionary");
+      error(errSyntaxError, -1, "Could not find catalog dictionary");
       catDict.free();
       return gFalse;
     }
@@ -268,7 +265,7 @@ GBool Catalog::cachePageTree(int page)
       obj.free();
     }
     else {
-      error(-1, "Top-level pages object is wrong type (%s)", obj.getTypeName());
+      error(errSyntaxError, -1, "Top-level pages object is wrong type ({0:s})", obj.getTypeName());
       obj.free();
       return gFalse;
     }
@@ -304,7 +301,7 @@ GBool Catalog::cachePageTree(int page)
     Object kids;
     pagesDict->lookup("Kids", &kids);
     if (!kids.isArray()) {
-      error(-1, "Kids object (page %d) is wrong type (%s)",
+      error(errSyntaxError, -1, "Kids object (page {0:d}) is wrong type ({1:s})",
             lastCachedPage+1, kids.getTypeName());
       kids.free();
       return gFalse;
@@ -328,7 +325,7 @@ GBool Catalog::cachePageTree(int page)
     Object kidRef;
     kids.arrayGetNF(kidsIdx, &kidRef);
     if (!kidRef.isRef()) {
-      error(-1, "Kid object (page %d) is not an indirect reference (%s)",
+      error(errSyntaxError, -1, "Kid object (page {0:d}) is not an indirect reference ({1:s})",
             lastCachedPage+1, kidRef.getTypeName());
       kidRef.free();
       kids.free();
@@ -343,7 +340,7 @@ GBool Catalog::cachePageTree(int page)
       }
     }
     if (loop) {
-      error(-1, "Loop in Pages tree");
+      error(errSyntaxError, -1, "Loop in Pages tree");
       kidRef.free();
       kids.free();
       kidsIdxList->back()++;
@@ -355,10 +352,10 @@ GBool Catalog::cachePageTree(int page)
     kids.free();
     if (kid.isDict("Page") || (kid.isDict() && !kid.getDict()->hasKey("Kids"))) {
       PageAttrs *attrs = new PageAttrs(attrsList->back(), kid.getDict());
-      Page *p = new Page(xref, lastCachedPage+1, kid.getDict(),
+      Page *p = new Page(doc, lastCachedPage+1, kid.getDict(),
                      kidRef.getRef(), attrs, form);
       if (!p->isOk()) {
-        error(-1, "Failed to create page (page %d)", lastCachedPage+1);
+        error(errSyntaxError, -1, "Failed to create page (page {0:d})", lastCachedPage+1);
         delete p;
         kidRef.free();
         kid.free();
@@ -366,7 +363,7 @@ GBool Catalog::cachePageTree(int page)
       }
 
       if (lastCachedPage >= numPages) {
-        error(-1, "Page count in top-level pages object is incorrect");
+        error(errSyntaxError, -1, "Page count in top-level pages object is incorrect");
         kidRef.free();
         kid.free();
         return gFalse;
@@ -388,7 +385,7 @@ GBool Catalog::cachePageTree(int page)
       pagesList->push_back(kid.getDict());
       kidsIdxList->push_back(0);
     } else {
-      error(-1, "Kid object (page %d) is wrong type (%s)",
+      error(errSyntaxError, -1, "Kid object (page {0:d}) is wrong type ({1:s})",
             lastCachedPage+1, kid.getTypeName());
       kidsIdxList->back()++;
     }
@@ -441,10 +438,10 @@ LinkDest *Catalog::findDest(GooString *name) {
     if (obj1.dictLookup("D", &obj2)->isArray())
       dest = new LinkDest(obj2.getArray());
     else
-      error(-1, "Bad named destination value");
+      error(errSyntaxWarning, -1, "Bad named destination value");
     obj2.free();
   } else {
-    error(-1, "Bad named destination value");
+    error(errSyntaxWarning, -1, "Bad named destination value");
   }
   obj1.free();
   if (dest && !dest->isOk()) {
@@ -465,6 +462,8 @@ FileSpec *Catalog::embeddedFile(int i)
       Object fsDict;
       embeddedFile = new FileSpec(obj.fetch(xref, &fsDict));
       fsDict.free();
+    } else if (obj.isDict()) {
+      embeddedFile = new FileSpec(&obj);
     } else {
       Object null;
       embeddedFile = new FileSpec(&null);
@@ -520,7 +519,7 @@ Catalog::PageMode Catalog::getPageMode() {
 
     xref->getCatalog(&catDict);
     if (!catDict.isDict()) {
-      error(-1, "Catalog object is wrong type (%s)", catDict.getTypeName());
+      error(errSyntaxError, -1, "Catalog object is wrong type ({0:s})", catDict.getTypeName());
       catDict.free();
       return pageMode;
     }
@@ -555,7 +554,7 @@ Catalog::PageLayout Catalog::getPageLayout() {
 
     xref->getCatalog(&catDict);
     if (!catDict.isDict()) {
-      error(-1, "Catalog object is wrong type (%s)", catDict.getTypeName());
+      error(errSyntaxError, -1, "Catalog object is wrong type ({0:s})", catDict.getTypeName());
       catDict.free();
       return pageLayout;
     }
@@ -607,7 +606,7 @@ NameTree::Entry::Entry(Array *array, int index) {
         name.append(aux.getString());
       }
       else
-        error(-1, "Invalid page tree");
+        error(errSyntaxError, -1, "Invalid page tree");
     }
 }
 
@@ -752,7 +751,7 @@ int Catalog::getNumPages()
 
     xref->getCatalog(&catDict);
     if (!catDict.isDict()) {
-      error(-1, "Catalog object is wrong type (%s)", catDict.getTypeName());
+      error(errSyntaxError, -1, "Catalog object is wrong type ({0:s})", catDict.getTypeName());
       catDict.free();
       return 0;
     }
@@ -762,7 +761,7 @@ int Catalog::getNumPages()
     // This should really be isDict("Pages"), but I've seen at least one
     // PDF file where the /Type entry is missing.
     if (!pagesDict.isDict()) {
-      error(-1, "Top-level pages object is wrong type (%s)",
+      error(errSyntaxError, -1, "Top-level pages object is wrong type ({0:s})",
           pagesDict.getTypeName());
       pagesDict.free();
       return 0;
@@ -771,7 +770,7 @@ int Catalog::getNumPages()
     pagesDict.dictLookup("Count", &obj);
     // some PDF files actually use real numbers here ("/Count 9.0")
     if (!obj.isNum()) {
-      error(-1, "Page count in top-level pages object is wrong type (%s)",
+      error(errSyntaxError, -1, "Page count in top-level pages object is wrong type ({0:s})",
          obj.getTypeName());
       numPages = 0;
     } else {
@@ -793,7 +792,7 @@ PageLabelInfo *Catalog::getPageLabelInfo()
 
     xref->getCatalog(&catDict);
     if (!catDict.isDict()) {
-      error(-1, "Catalog object is wrong type (%s)", catDict.getTypeName());
+      error(errSyntaxError, -1, "Catalog object is wrong type ({0:s})", catDict.getTypeName());
       catDict.free();
       return NULL;
     }
@@ -818,7 +817,7 @@ Object *Catalog::getStructTreeRoot()
      if (catDict.isDict()) {
        catDict.dictLookup("StructTreeRoot", &structTreeRoot);
      } else {
-       error(-1, "Catalog object is wrong type (%s)", catDict.getTypeName());
+       error(errSyntaxError, -1, "Catalog object is wrong type ({0:s})", catDict.getTypeName());
        structTreeRoot.initNull();
      }
      catDict.free();
@@ -837,7 +836,7 @@ Object *Catalog::getOutline()
      if (catDict.isDict()) {
        catDict.dictLookup("Outlines", &outline);
      } else {
-       error(-1, "Catalog object is wrong type (%s)", catDict.getTypeName());
+       error(errSyntaxError, -1, "Catalog object is wrong type ({0:s})", catDict.getTypeName());
        outline.initNull();
      }
      catDict.free();
@@ -856,7 +855,7 @@ Object *Catalog::getDests()
      if (catDict.isDict()) {
        catDict.dictLookup("Dests", &dests);
      } else {
-       error(-1, "Catalog object is wrong type (%s)", catDict.getTypeName());
+       error(errSyntaxError, -1, "Catalog object is wrong type ({0:s})", catDict.getTypeName());
        dests.initNull();
      }
      catDict.free();
@@ -869,7 +868,9 @@ Form *Catalog::getForm()
 {
   if (!form) {
     if (acroForm.isDict()) {
-      form = new Form(xref,&acroForm);
+      form = new Form(doc, &acroForm);
+      // perform form-related loading after all widgets have been loaded
+      form->postWidgetsLoad();
     }
   }
 
@@ -897,7 +898,7 @@ Object *Catalog::getNames()
      if (catDict.isDict()) {
        catDict.dictLookup("Names", &names);
      } else {
-       error(-1, "Catalog object is wrong type (%s)", catDict.getTypeName());
+       error(errSyntaxError, -1, "Catalog object is wrong type ({0:s})", catDict.getTypeName());
        names.initNull();
      }
      catDict.free();
