@@ -17,6 +17,7 @@
 // Copyright (C) 2006 Jeff Muizelaar <jeff@infidigm.net>
 // Copyright (C) 2010 Christian Feuersnger <cfeuersaenger@googlemail.com>
 // Copyright (C) 2011 Andrea Canciani <ranma42@gmail.com>
+// Copyright (C) 2012 Thomas Freitag <Thomas.Freitag@alfa.de>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -73,12 +74,12 @@ Function *Function::parse(Object *funcObj, std::set<int> *usedParents) {
   } else if (funcObj->isName("Identity")) {
     return new IdentityFunction();
   } else {
-    error(-1, "Expected function dictionary or stream");
+    error(errSyntaxError, -1, "Expected function dictionary or stream");
     return NULL;
   }
 
   if (!dict->lookup("FunctionType", &obj1)->isInt()) {
-    error(-1, "Function type is missing or wrong type");
+    error(errSyntaxError, -1, "Function type is missing or wrong type");
     obj1.free();
     return NULL;
   }
@@ -94,7 +95,7 @@ Function *Function::parse(Object *funcObj, std::set<int> *usedParents) {
   } else if (funcType == 4) {
     func = new PostScriptFunction(funcObj, dict);
   } else {
-    error(-1, "Unimplemented function type (%d)", funcType);
+    error(errSyntaxError, -1, "Unimplemented function type ({0:d})", funcType);
     return NULL;
   }
   if (!func->isOk()) {
@@ -111,26 +112,26 @@ GBool Function::init(Dict *dict) {
 
   //----- Domain
   if (!dict->lookup("Domain", &obj1)->isArray()) {
-    error(-1, "Function is missing domain");
+    error(errSyntaxError, -1, "Function is missing domain");
     goto err2;
   }
   m = obj1.arrayGetLength() / 2;
   if (m > funcMaxInputs) {
-    error(-1, "Functions with more than %d inputs are unsupported",
+    error(errSyntaxError, -1, "Functions with more than {0:d} inputs are unsupported",
 	  funcMaxInputs);
     goto err2;
   }
   for (i = 0; i < m; ++i) {
     obj1.arrayGet(2*i, &obj2);
     if (!obj2.isNum()) {
-      error(-1, "Illegal value in function domain array");
+      error(errSyntaxError, -1, "Illegal value in function domain array");
       goto err1;
     }
     domain[i][0] = obj2.getNum();
     obj2.free();
     obj1.arrayGet(2*i+1, &obj2);
     if (!obj2.isNum()) {
-      error(-1, "Illegal value in function domain array");
+      error(errSyntaxError, -1, "Illegal value in function domain array");
       goto err1;
     }
     domain[i][1] = obj2.getNum();
@@ -145,21 +146,21 @@ GBool Function::init(Dict *dict) {
     hasRange = gTrue;
     n = obj1.arrayGetLength() / 2;
     if (n > funcMaxOutputs) {
-      error(-1, "Functions with more than %d outputs are unsupported",
+      error(errSyntaxError, -1, "Functions with more than {0:d} outputs are unsupported",
 	    funcMaxOutputs);
       goto err2;
     }
     for (i = 0; i < n; ++i) {
       obj1.arrayGet(2*i, &obj2);
       if (!obj2.isNum()) {
-	error(-1, "Illegal value in function range array");
+	error(errSyntaxError, -1, "Illegal value in function range array");
 	goto err1;
       }
       range[i][0] = obj2.getNum();
       obj2.free();
       obj1.arrayGet(2*i+1, &obj2);
       if (!obj2.isNum()) {
-	error(-1, "Illegal value in function range array");
+	error(errSyntaxError, -1, "Illegal value in function range array");
 	goto err1;
       }
       range[i][1] = obj2.getNum();
@@ -218,8 +219,10 @@ SampledFunction::SampledFunction(Object *funcObj, Dict *dict) {
   Guint buf, bitMask;
   int bits;
   Guint s;
-  int i;
+  double in[funcMaxInputs];
+  int i, j, t, bit, idx;
 
+  idxOffset = NULL;
   samples = NULL;
   sBuf = NULL;
   ok = gFalse;
@@ -229,11 +232,11 @@ SampledFunction::SampledFunction(Object *funcObj, Dict *dict) {
     goto err1;
   }
   if (!hasRange) {
-    error(-1, "Type 0 function is missing range");
+    error(errSyntaxError, -1, "Type 0 function is missing range");
     goto err1;
   }
   if (m > sampledFuncMaxInputs) {
-    error(-1, "Sampled functions with more than %d inputs are unsupported",
+    error(errSyntaxError, -1, "Sampled functions with more than {0:d} inputs are unsupported",
 	  sampledFuncMaxInputs);
     goto err1;
   }
@@ -243,7 +246,7 @@ SampledFunction::SampledFunction(Object *funcObj, Dict *dict) {
 
   //----- get the stream
   if (!funcObj->isStream()) {
-    error(-1, "Type 0 function isn't a stream");
+    error(errSyntaxError, -1, "Type 0 function isn't a stream");
     goto err1;
   }
   str = funcObj->getStream();
@@ -251,27 +254,45 @@ SampledFunction::SampledFunction(Object *funcObj, Dict *dict) {
   //----- Size
   if (!dict->lookup("Size", &obj1)->isArray() ||
       obj1.arrayGetLength() != m) {
-    error(-1, "Function has missing or invalid size array");
+    error(errSyntaxError, -1, "Function has missing or invalid size array");
     goto err2;
   }
   for (i = 0; i < m; ++i) {
     obj1.arrayGet(i, &obj2);
     if (!obj2.isInt()) {
-      error(-1, "Illegal value in function size array");
+      error(errSyntaxError, -1, "Illegal value in function size array");
       goto err3;
     }
     sampleSize[i] = obj2.getInt();
+    if (sampleSize[i] <= 0) {
+      error(errSyntaxError, -1, "Illegal non-positive value in function size array");
+      goto err3;
+    }
     obj2.free();
   }
   obj1.free();
-  idxMul[0] = n;
-  for (i = 1; i < m; ++i) {
-    idxMul[i] = idxMul[i-1] * sampleSize[i-1];
+  idxOffset = (int *)gmallocn(1 << m, sizeof(int));
+  for (i = 0; i < (1<<m); ++i) {
+    idx = 0;
+    for (j = m - 1, t = i; j >= 1; --j, t <<= 1) {
+      if (sampleSize[j] == 1) {
+	bit = 0;
+      } else {
+	bit = (t >> (m - 1)) & 1;
+      }
+      idx = (idx + bit) * sampleSize[j-1];
+    }
+    if (sampleSize[0] == 1) {
+      bit = 0;
+    } else {
+      bit = (t >> (m - 1)) & 1;
+    }
+    idxOffset[i] = (idx + bit) * n;
   }
 
   //----- BitsPerSample
   if (!dict->lookup("BitsPerSample", &obj1)->isInt()) {
-    error(-1, "Function has missing or invalid BitsPerSample");
+    error(errSyntaxError, -1, "Function has missing or invalid BitsPerSample");
     goto err2;
   }
   sampleBits = obj1.getInt();
@@ -284,14 +305,14 @@ SampledFunction::SampledFunction(Object *funcObj, Dict *dict) {
     for (i = 0; i < m; ++i) {
       obj1.arrayGet(2*i, &obj2);
       if (!obj2.isNum()) {
-	error(-1, "Illegal value in function encode array");
+	error(errSyntaxError, -1, "Illegal value in function encode array");
 	goto err3;
       }
       encode[i][0] = obj2.getNum();
       obj2.free();
       obj1.arrayGet(2*i+1, &obj2);
       if (!obj2.isNum()) {
-	error(-1, "Illegal value in function encode array");
+	error(errSyntaxError, -1, "Illegal value in function encode array");
 	goto err3;
       }
       encode[i][1] = obj2.getNum();
@@ -315,14 +336,14 @@ SampledFunction::SampledFunction(Object *funcObj, Dict *dict) {
     for (i = 0; i < n; ++i) {
       obj1.arrayGet(2*i, &obj2);
       if (!obj2.isNum()) {
-	error(-1, "Illegal value in function decode array");
+	error(errSyntaxError, -1, "Illegal value in function decode array");
 	goto err3;
       }
       decode[i][0] = obj2.getNum();
       obj2.free();
       obj1.arrayGet(2*i+1, &obj2);
       if (!obj2.isNum()) {
-	error(-1, "Illegal value in function decode array");
+	error(errSyntaxError, -1, "Illegal value in function decode array");
 	goto err3;
       }
       decode[i][1] = obj2.getNum();
@@ -368,6 +389,13 @@ SampledFunction::SampledFunction(Object *funcObj, Dict *dict) {
   }
   str->close();
 
+  // set up the cache
+  for (i = 0; i < m; ++i) {
+    in[i] = domain[i][0];
+    cacheIn[i] = in[i] - 1;
+  }
+  transform(in, cacheOut);
+
   ok = gTrue;
   return;
 
@@ -380,6 +408,9 @@ SampledFunction::SampledFunction(Object *funcObj, Dict *dict) {
 }
 
 SampledFunction::~SampledFunction() {
+  if (idxOffset) {
+    gfree(idxOffset);
+  }
   if (samples) {
     gfree(samples);
   }
@@ -390,6 +421,8 @@ SampledFunction::~SampledFunction() {
 
 SampledFunction::SampledFunction(SampledFunction *func) {
   memcpy(this, func, sizeof(SampledFunction));
+  idxOffset = (int *)gmallocn(1 << m, sizeof(int));
+  memcpy(idxOffset, func->idxOffset, (1 << m) * (int)sizeof(int));
   samples = (double *)gmallocn(nSamples, sizeof(double));
   memcpy(samples, func->samples, nSamples * sizeof(double));
   sBuf = (double *)gmallocn(1 << m, sizeof(double));
@@ -397,37 +430,54 @@ SampledFunction::SampledFunction(SampledFunction *func) {
 
 void SampledFunction::transform(double *in, double *out) {
   double x;
-  int e[funcMaxInputs][2];
+  int e[funcMaxInputs];
   double efrac0[funcMaxInputs];
   double efrac1[funcMaxInputs];
-  int i, j, k, idx, t;
+  int i, j, k, idx0, t;
+
+  // check the cache
+  for (i = 0; i < m; ++i) {
+    if (in[i] != cacheIn[i]) {
+      break;
+    }
+  }
+  if (i == m) {
+    for (i = 0; i < n; ++i) {
+      out[i] = cacheOut[i];
+    }
+    return;
+  }
 
   // map input values into sample array
   for (i = 0; i < m; ++i) {
     x = (in[i] - domain[i][0]) * inputMul[i] + encode[i][0];
-    if (x < 0) {
+    if (x < 0 || x != x) {  // x!=x is a more portable version of isnan(x)
       x = 0;
     } else if (x > sampleSize[i] - 1) {
       x = sampleSize[i] - 1;
     }
-    e[i][0] = (int)x;
-    if ((e[i][1] = e[i][0] + 1) >= sampleSize[i]) {
+    e[i] = (int)x;
+    if (e[i] == sampleSize[i] - 1 && sampleSize[i] > 1) {
       // this happens if in[i] = domain[i][1]
-      e[i][1] = e[i][0];
+      e[i] = sampleSize[i] - 2;
     }
-    efrac1[i] = x - e[i][0];
+    efrac1[i] = x - e[i];
     efrac0[i] = 1 - efrac1[i];
   }
 
+  // compute index for the first sample to be used
+  idx0 = 0;
+  for (k = m - 1; k >= 1; --k) {
+    idx0 = (idx0 + e[k]) * sampleSize[k-1];
+  }
+  idx0 = (idx0 + e[0]) * n;
+
   // for each output, do m-linear interpolation
   for (i = 0; i < n; ++i) {
 
     // pull 2^m values out of the sample array
     for (j = 0; j < (1<<m); ++j) {
-      idx = i;
-      for (k = 0, t = j; k < m; ++k, t >>= 1) {
-	idx += idxMul[k] * (e[k][t & 1]);
-      }
+      int idx = idx0 + idxOffset[j] + i;
       if (likely(idx >= 0 && idx < nSamples)) {
         sBuf[j] = samples[idx];
       } else {
@@ -450,6 +500,14 @@ void SampledFunction::transform(double *in, double *out) {
       out[i] = range[i][1];
     }
   }
+
+  // save current result in the cache
+  for (i = 0; i < m; ++i) {
+    cacheIn[i] = in[i];
+  }
+  for (i = 0; i < n; ++i) {
+    cacheOut[i] = out[i];
+  }
 }
 
 //------------------------------------------------------------------------
@@ -467,21 +525,21 @@ ExponentialFunction::ExponentialFunction(Object *funcObj, Dict *dict) {
     goto err1;
   }
   if (m != 1) {
-    error(-1, "Exponential function with more than one input");
+    error(errSyntaxError, -1, "Exponential function with more than one input");
     goto err1;
   }
 
   //----- C0
   if (dict->lookup("C0", &obj1)->isArray()) {
     if (hasRange && obj1.arrayGetLength() != n) {
-      error(-1, "Function's C0 array is wrong length");
+      error(errSyntaxError, -1, "Function's C0 array is wrong length");
       goto err2;
     }
     n = obj1.arrayGetLength();
     for (i = 0; i < n; ++i) {
       obj1.arrayGet(i, &obj2);
       if (!obj2.isNum()) {
-	error(-1, "Illegal value in function C0 array");
+	error(errSyntaxError, -1, "Illegal value in function C0 array");
 	goto err3;
       }
       c0[i] = obj2.getNum();
@@ -489,7 +547,7 @@ ExponentialFunction::ExponentialFunction(Object *funcObj, Dict *dict) {
     }
   } else {
     if (hasRange && n != 1) {
-      error(-1, "Function's C0 array is wrong length");
+      error(errSyntaxError, -1, "Function's C0 array is wrong length");
       goto err2;
     }
     n = 1;
@@ -500,13 +558,13 @@ ExponentialFunction::ExponentialFunction(Object *funcObj, Dict *dict) {
   //----- C1
   if (dict->lookup("C1", &obj1)->isArray()) {
     if (obj1.arrayGetLength() != n) {
-      error(-1, "Function's C1 array is wrong length");
+      error(errSyntaxError, -1, "Function's C1 array is wrong length");
       goto err2;
     }
     for (i = 0; i < n; ++i) {
       obj1.arrayGet(i, &obj2);
       if (!obj2.isNum()) {
-	error(-1, "Illegal value in function C1 array");
+	error(errSyntaxError, -1, "Illegal value in function C1 array");
 	goto err3;
       }
       c1[i] = obj2.getNum();
@@ -514,7 +572,7 @@ ExponentialFunction::ExponentialFunction(Object *funcObj, Dict *dict) {
     }
   } else {
     if (n != 1) {
-      error(-1, "Function's C1 array is wrong length");
+      error(errSyntaxError, -1, "Function's C1 array is wrong length");
       goto err2;
     }
     c1[0] = 1;
@@ -523,7 +581,7 @@ ExponentialFunction::ExponentialFunction(Object *funcObj, Dict *dict) {
 
   //----- N (exponent)
   if (!dict->lookup("N", &obj1)->isNum()) {
-    error(-1, "Function has missing or invalid N");
+    error(errSyntaxError, -1, "Function has missing or invalid N");
     goto err2;
   }
   e = obj1.getNum();
@@ -591,13 +649,13 @@ StitchingFunction::StitchingFunction(Object *funcObj, Dict *dict, std::set<int>
     goto err1;
   }
   if (m != 1) {
-    error(-1, "Stitching function with more than one input");
+    error(errSyntaxError, -1, "Stitching function with more than one input");
     goto err1;
   }
 
   //----- Functions
   if (!dict->lookup("Functions", &obj1)->isArray()) {
-    error(-1, "Missing 'Functions' entry in stitching function");
+    error(errSyntaxError, -1, "Missing 'Functions' entry in stitching function");
     goto err1;
   }
   k = obj1.arrayGetLength();
@@ -624,9 +682,10 @@ StitchingFunction::StitchingFunction(Object *funcObj, Dict *dict, std::set<int>
     if (!(funcs[i] = Function::parse(&obj2, &usedParentsAux))) {
       goto err2;
     }
-    if (i > 0 && (funcs[i]->getInputSize() != 1 ||
-		  funcs[i]->getOutputSize() != funcs[0]->getOutputSize())) {
-      error(-1, "Incompatible subfunctions in stitching function");
+    if (funcs[i]->getInputSize() != 1 ||
+	(i > 0 && funcs[i]->getOutputSize() != funcs[0]->getOutputSize())) {
+      error(errSyntaxError, -1,
+	    "Incompatible subfunctions in stitching function");
       goto err2;
     }
     obj2.free();
@@ -636,13 +695,13 @@ StitchingFunction::StitchingFunction(Object *funcObj, Dict *dict, std::set<int>
   //----- Bounds
   if (!dict->lookup("Bounds", &obj1)->isArray() ||
       obj1.arrayGetLength() != k - 1) {
-    error(-1, "Missing or invalid 'Bounds' entry in stitching function");
+    error(errSyntaxError, -1, "Missing or invalid 'Bounds' entry in stitching function");
     goto err1;
   }
   bounds[0] = domain[0][0];
   for (i = 1; i < k; ++i) {
     if (!obj1.arrayGet(i - 1, &obj2)->isNum()) {
-      error(-1, "Invalid type in 'Bounds' array in stitching function");
+      error(errSyntaxError, -1, "Invalid type in 'Bounds' array in stitching function");
       goto err2;
     }
     bounds[i] = obj2.getNum();
@@ -654,12 +713,12 @@ StitchingFunction::StitchingFunction(Object *funcObj, Dict *dict, std::set<int>
   //----- Encode
   if (!dict->lookup("Encode", &obj1)->isArray() ||
       obj1.arrayGetLength() != 2 * k) {
-    error(-1, "Missing or invalid 'Encode' entry in stitching function");
+    error(errSyntaxError, -1, "Missing or invalid 'Encode' entry in stitching function");
     goto err1;
   }
   for (i = 0; i < 2 * k; ++i) {
     if (!obj1.arrayGet(i, &obj2)->isNum()) {
-      error(-1, "Invalid type in 'Encode' array in stitching function");
+      error(errSyntaxError, -1, "Invalid type in 'Encode' array in stitching function");
       goto err2;
     }
     encode[i] = obj2.getNum();
@@ -692,8 +751,6 @@ StitchingFunction::StitchingFunction(StitchingFunction *func) {
   int i;
 
   memcpy(this, func, sizeof(StitchingFunction));
-
-  k = func->k;
   funcs = (Function **)gmallocn(k, sizeof(Function *));
   for (i = 0; i < k; ++i) {
     funcs[i] = func->funcs[i]->copy();
@@ -954,6 +1011,10 @@ public:
       return;
     }
     --sp;
+    if (sp + i + 1 >= psStackSize) {
+      error(errSyntaxError, -1, "Stack underflow in PostScript function");
+      return;
+    }
     stack[sp] = stack[sp + 1 + i];
   }
   void pop()
@@ -969,7 +1030,7 @@ private:
   GBool checkOverflow(int n = 1)
   {
     if (sp - n < 0) {
-      error(-1, "Stack overflow in PostScript function");
+      error(errSyntaxError, -1, "Stack overflow in PostScript function");
       return gFalse;
     }
     return gTrue;
@@ -977,7 +1038,7 @@ private:
   GBool checkUnderflow()
   {
     if (sp == psStackSize) {
-      error(-1, "Stack underflow in PostScript function");
+      error(errSyntaxError, -1, "Stack underflow in PostScript function");
       return gFalse;
     }
     return gTrue;
@@ -985,7 +1046,7 @@ private:
   GBool checkType(PSObjectType t1, PSObjectType t2)
   {
     if (stack[sp].type != t1 && stack[sp].type != t2) {
-      error(-1, "Type mismatch in PostScript function");
+      error(errSyntaxError, -1, "Type mismatch in PostScript function");
       return gFalse;
     }
     return gTrue;
@@ -999,7 +1060,7 @@ void PSStack::copy(int n) {
   int i;
 
   if (sp + n > psStackSize) {
-    error(-1, "Stack underflow in PostScript function");
+    error(errSyntaxError, -1, "Stack underflow in PostScript function");
     return;
   }
   if (!checkOverflow(n)) {
@@ -1026,7 +1087,7 @@ void PSStack::roll(int n, int j) {
       j = n - j;
     }
   }
-  if (n <= 0 || j == 0) {
+  if (n <= 0 || j == 0 || n > psStackSize || sp + n > psStackSize) {
     return;
   }
   if (j <= n / 2) {
@@ -1053,6 +1114,8 @@ PostScriptFunction::PostScriptFunction(Object *funcObj, Dict *dict) {
   Stream *str;
   int codePtr;
   GooString *tok;
+  double in[funcMaxInputs];
+  int i;
 
   code = NULL;
   codeString = NULL;
@@ -1064,13 +1127,13 @@ PostScriptFunction::PostScriptFunction(Object *funcObj, Dict *dict) {
     goto err1;
   }
   if (!hasRange) {
-    error(-1, "Type 4 function is missing range");
+    error(errSyntaxError, -1, "Type 4 function is missing range");
     goto err1;
   }
 
   //----- get the stream
   if (!funcObj->isStream()) {
-    error(-1, "Type 4 function isn't a stream");
+    error(errSyntaxError, -1, "Type 4 function isn't a stream");
     goto err1;
   }
   str = funcObj->getStream();
@@ -1079,7 +1142,7 @@ PostScriptFunction::PostScriptFunction(Object *funcObj, Dict *dict) {
   codeString = new GooString();
   str->reset();
   if (!(tok = getToken(str)) || tok->cmp("{")) {
-    error(-1, "Expected '{' at start of PostScript function");
+    error(errSyntaxError, -1, "Expected '{' at start of PostScript function");
     if (tok) {
       delete tok;
     }
@@ -1092,6 +1155,13 @@ PostScriptFunction::PostScriptFunction(Object *funcObj, Dict *dict) {
   }
   str->close();
 
+  //----- set up the cache
+  for (i = 0; i < m; ++i) {
+    in[i] = domain[i][0];
+    cacheIn[i] = in[i] - 1;
+  }
+  transform(in, cacheOut);
+
   ok = gTrue;
   
  err2:
@@ -1115,7 +1185,20 @@ PostScriptFunction::~PostScriptFunction() {
 void PostScriptFunction::transform(double *in, double *out) {
   PSStack stack;
   int i;
-  
+
+  // check the cache
+  for (i = 0; i < m; ++i) {
+    if (in[i] != cacheIn[i]) {
+      break;
+    }
+  }
+  if (i == m) {
+    for (i = 0; i < n; ++i) {
+      out[i] = cacheOut[i];
+    }
+    return;
+  }
+
   for (i = 0; i < m; ++i) {
     //~ may need to check for integers here
     stack.pushReal(in[i]);
@@ -1131,9 +1214,18 @@ void PostScriptFunction::transform(double *in, double *out) {
   }
   stack.clear();
 
-  // if (!stack.empty()) {
-  //   error(-1, "Extra values on stack at end of PostScript function");
+  // if (!stack->empty()) {
+  //   error(errSyntaxWarning, -1,
+  //         "Extra values on stack at end of PostScript function");
   // }
+
+  // save current result in the cache
+  for (i = 0; i < m; ++i) {
+    cacheIn[i] = in[i];
+  }
+  for (i = 0; i < n; ++i) {
+    cacheOut[i] = out[i];
+  }
 }
 
 GBool PostScriptFunction::parseCode(Stream *str, int *codePtr) {
@@ -1145,13 +1237,13 @@ GBool PostScriptFunction::parseCode(Stream *str, int *codePtr) {
 
   while (1) {
     if (!(tok = getToken(str))) {
-      error(-1, "Unexpected end of PostScript function stream");
+      error(errSyntaxError, -1, "Unexpected end of PostScript function stream");
       return gFalse;
     }
     p = tok->getCString();
     if (isdigit(*p) || *p == '.' || *p == '-') {
       isReal = gFalse;
-      for (++p; *p; ++p) {
+      for (; *p; ++p) {
 	if (*p == '.') {
 	  isReal = gTrue;
 	  break;
@@ -1176,7 +1268,7 @@ GBool PostScriptFunction::parseCode(Stream *str, int *codePtr) {
 	return gFalse;
       }
       if (!(tok = getToken(str))) {
-	error(-1, "Unexpected end of PostScript function stream");
+	error(errSyntaxError, -1, "Unexpected end of PostScript function stream");
 	return gFalse;
       }
       if (!tok->cmp("{")) {
@@ -1186,7 +1278,7 @@ GBool PostScriptFunction::parseCode(Stream *str, int *codePtr) {
 	}
 	delete tok;
 	if (!(tok = getToken(str))) {
-	  error(-1, "Unexpected end of PostScript function stream");
+	  error(errSyntaxError, -1, "Unexpected end of PostScript function stream");
 	  return gFalse;
 	}
       } else {
@@ -1194,7 +1286,8 @@ GBool PostScriptFunction::parseCode(Stream *str, int *codePtr) {
       }
       if (!tok->cmp("if")) {
 	if (elsePtr >= 0) {
-	  error(-1, "Got 'if' operator with two blocks in PostScript function");
+	  error(errSyntaxError, -1,
+		"Got 'if' operator with two blocks in PostScript function");
 	  return gFalse;
 	}
 	code[opPtr].type = psOperator;
@@ -1203,7 +1296,8 @@ GBool PostScriptFunction::parseCode(Stream *str, int *codePtr) {
 	code[opPtr+2].blk = *codePtr;
       } else if (!tok->cmp("ifelse")) {
 	if (elsePtr < 0) {
-	  error(-1, "Got 'ifelse' operator with one blocks in PostScript function");
+	  error(errSyntaxError, -1,
+		"Got 'ifelse' operator with one block in PostScript function");
 	  return gFalse;
 	}
 	code[opPtr].type = psOperator;
@@ -1213,7 +1307,8 @@ GBool PostScriptFunction::parseCode(Stream *str, int *codePtr) {
 	code[opPtr+2].type = psBlock;
 	code[opPtr+2].blk = *codePtr;
       } else {
-	error(-1, "Expected if/ifelse operator in PostScript function");
+	error(errSyntaxError, -1,
+	      "Expected if/ifelse operator in PostScript function");
 	delete tok;
 	return gFalse;
       }
@@ -1228,6 +1323,7 @@ GBool PostScriptFunction::parseCode(Stream *str, int *codePtr) {
     } else {
       a = -1;
       b = nPSOps;
+      cmp = 0; // make gcc happy
       // invariant: psOpNames[a] < tok < psOpNames[b]
       while (b - a > 1) {
 	mid = (a + b) / 2;
@@ -1241,8 +1337,9 @@ GBool PostScriptFunction::parseCode(Stream *str, int *codePtr) {
 	}
       }
       if (cmp != 0) {
-	error(-1, "Unknown operator '%s' in PostScript function",
-	      tok->getCString());
+	error(errSyntaxError, -1,
+	      "Unknown operator '{0:t}' in PostScript function",
+	      tok);
 	delete tok;
 	return gFalse;
       }
@@ -1368,7 +1465,7 @@ void PostScriptFunction::exec(PSStack *stack, int codePtr) {
 	if (i2 > 0) {
 	  stack->pushInt(i1 << i2);
 	} else if (i2 < 0) {
-	  stack->pushInt((int)((Guint)i1 >> i2));
+	  stack->pushInt((int)((Guint)i1 >> -i2));
 	} else {
 	  stack->pushInt(i1);
 	}
@@ -1620,7 +1717,7 @@ void PostScriptFunction::exec(PSStack *stack, int codePtr) {
       }
       break;
     default:
-      error(-1, "Internal: bad object in PostScript function code");
+      error(errSyntaxError, -1, "Internal: bad object in PostScript function code");
       break;
     }
   }
