@@ -20,8 +20,10 @@
 #pragma comment(linker,"/manifestdependency:\"type='win32' name='Microsoft.Windows.Common-Controls' version='6.0.0.0' processorArchitecture='x86' publicKeyToken='6595b64144ccf1df' language='*'\"")
 #endif
 
-#ifdef WIN32
+#ifdef _WIN32
 #include <windows.h>
+#else
+#include <strings.h>
 #endif
 
 // Define COPY_FILE if you want the file to be copied to a local disk first
@@ -297,7 +299,7 @@ bool str_endswith(const char *txt, const char *end)
    sleep_milliseconds */
 void sleep_milliseconds(int milliseconds)
 {
-#ifdef WIN32
+#ifdef _WIN32
     Sleep((DWORD)milliseconds);
 #else
     struct timespec tv;
@@ -429,7 +431,7 @@ SplashOutputDev * PdfEnginePoppler::outputDevice() {
         GBool bitmapTopDown = gTrue;
         _outputDev = new SplashOutputDev(gSplashColorMode, 4, gFalse, gBgColor, bitmapTopDown);
         if (_outputDev)
-            _outputDev->startDoc(_pdfDoc->getXRef());
+            _outputDev->startDoc(_pdfDoc);
     }
     return _outputDev;
 }
@@ -456,7 +458,7 @@ struct FindFileState {
     char dirpath[MAX_FILENAME_SIZE]; /* current dir path */
     char pattern[MAX_FILENAME_SIZE]; /* search pattern */
     const char *bufptr;
-#ifdef WIN32
+#ifdef _WIN32
     WIN32_FIND_DATA fileinfo;
     HANDLE dir;
 #else
@@ -464,7 +466,7 @@ struct FindFileState {
 #endif
 };
 
-#ifdef WIN32
+#ifdef _WIN32
 #include <windows.h>
 #include <sys/timeb.h>
 #include <direct.h>
@@ -495,7 +497,7 @@ int fnmatch(const char *pattern, const char *string, int flags)
 #include <fnmatch.h>
 #endif
 
-#ifdef WIN32
+#ifdef _WIN32
 /* on windows to query dirs we need foo\* to get files in this directory.
     foo\ always fails and foo will return just info about foo directory,
     not files in this directory */
@@ -520,12 +522,12 @@ FindFileState *find_file_open(const char *path, const char *pattern)
         return NULL;
     strcpy_s(s->path, sizeof(s->path), path);
     strcpy_s(s->dirpath, sizeof(s->path), path);
-#ifdef WIN32
+#ifdef _WIN32
     win_correct_path_for_FindFirstFile(s->path, sizeof(s->path));
 #endif
     strcpy_s(s->pattern, sizeof(s->pattern), pattern);
     s->bufptr = s->path;
-#ifdef WIN32
+#ifdef _WIN32
     s->dir = INVALID_HANDLE_VALUE;
 #else
     s->dir = NULL;
@@ -553,7 +555,7 @@ char *makepath(char *buf, int buf_size, const char *path,
     return buf;
 }
 
-#ifdef WIN32
+#ifdef _WIN32
 static int skip_matching_file(const char *filename)
 {
     if (0 == strcmp(".", filename))
@@ -566,7 +568,7 @@ static int skip_matching_file(const char *filename)
 
 int find_file_next(FindFileState *s, char *filename, int filename_size_max)
 {
-#ifdef WIN32
+#ifdef _WIN32
     int    fFound;
     if (INVALID_HANDLE_VALUE == s->dir) {
         s->dir = FindFirstFile(s->path, &(s->fileinfo));
@@ -633,7 +635,7 @@ CheckFile:
 
 void find_file_close(FindFileState *s)
 {
-#ifdef WIN32
+#ifdef _WIN32
     if (INVALID_HANDLE_VALUE != s->dir)
        FindClose(s->dir);
 #else
@@ -731,7 +733,7 @@ void StrList_Destroy(StrList **root)
     *root = NULL;
 }
 
-#ifndef WIN32
+#ifndef _WIN32
 void OutputDebugString(const char *txt)
 {
     /* do nothing */
@@ -740,7 +742,7 @@ void OutputDebugString(const char *txt)
 #define _vsnprintf vsnprintf
 #endif
 
-void my_error(int pos, char *msg, va_list args) {
+void my_error(void *, ErrorCategory, int pos, char *msg) {
 #if 0
     char        buf[4096], *p = buf;
 
@@ -794,7 +796,7 @@ void my_error(int pos, char *msg, va_list args) {
 #endif
 }
 
-void LogInfo(char *fmt, ...)
+void LogInfo(const char *fmt, ...)
 {
     va_list args;
     char        buf[4096], *p = buf;
@@ -838,7 +840,7 @@ static void RenderPdfAsText(const char *fileName)
 
     LogInfo("started: %s\n", fileName);
 
-    TextOutputDev * textOut = new TextOutputDev(NULL, gTrue, gFalse, gFalse);
+    TextOutputDev * textOut = new TextOutputDev(NULL, gTrue, 0, gFalse, gFalse);
     if (!textOut->isOk()) {
         delete textOut;
         return;
@@ -852,7 +854,7 @@ static void RenderPdfAsText(const char *fileName)
 
     pdfDoc = new PDFDoc(fileNameStr, NULL, NULL, NULL);
     if (!pdfDoc->isOk()) {
-        error(-1, "RenderPdfFile(): failed to open PDF file %s\n", fileName);
+        error(errIO, -1, "RenderPdfFile(): failed to open PDF file {0:s}\n", fileName);
         goto Exit;
     }
 
@@ -1120,7 +1122,7 @@ Exit:
 }
 #endif
 
-#ifdef WIN32
+#ifdef _WIN32
 #include <sys/types.h>
 #include <sys/stat.h>
 
@@ -1223,13 +1225,13 @@ static void RenderCmdLineArg(char *cmdLineArg)
             RenderFileList(cmdLineArg);
 #endif
     } else {
-        error(-1, "unexpected argument '%s'", cmdLineArg);
+        error(errCommandLine, -1, "unexpected argument '{0:s}'", cmdLineArg);
     }
 }
 
 int main(int argc, char **argv)
 {
-    setErrorFunction(my_error);
+    setErrorCallback(my_error, NULL);
     ParseCommandLine(argc, argv);
     if (0 == StrList_Len(&gArgsListRoot))
         PrintUsageAndExit(argc, argv);
