@@ -6,6 +6,7 @@
 //
 // Copyright (C) 2011 Thomas Freitag <Thomas.Freitag@alfa.de>
 // Copyright (C) 2012 Arseny Solokha <asolokha@gmx.com>
+// Copyright (C) 2012 Fabio D'Urso <fabiodurso@hotmail.it>
 //
 //========================================================================
 #include <PDFDoc.h>
@@ -84,16 +85,16 @@ int main (int argc, char *argv[])
         }
       }
     } else if (doc->isOk()) {
-      error(-1, "Could not merge encrypted files ('%s')", argv[i]);
+      error(errUnimplemented, -1, "Could not merge encrypted files ('{0:s}')", argv[i]);
       return -1;
     } else {
-      error(-1, "Could not merge damaged documents ('%s')", argv[i]);
+      error(errSyntaxError, -1, "Could not merge damaged documents ('{0:s}')", argv[i]);
       return -1;
     }
   }
 
   if (!(f = fopen(fileName, "wb"))) {
-    error(-1, "Could not open file '%s'", fileName);
+    error(errIO, -1, "Could not open file '{0:s}'", fileName);
     return -1;
   }
   outStr = new FileOutStream(f, 0);
@@ -161,13 +162,14 @@ int main (int argc, char *argv[])
     objectsCount++;
   }
   Guint uxrefOffset = outStr->getPos();
-  yRef->writeToFile(outStr, gFalse /* do not write unnecessary entries */ );
-
   Ref ref;
   ref.num = rootNum;
   ref.gen = 0;
-  PDFDoc::writeTrailer(uxrefOffset, objectsCount, outStr, (GBool) gFalse, 0,
-	&ref, yRef, fileName, outStr->getPos());
+  Dict *trailerDict = PDFDoc::createTrailerDict(objectsCount, gFalse, 0, &ref, yRef,
+                                                fileName, outStr->getPos());
+  PDFDoc::writeXRefTableTrailer(trailerDict, yRef, gFalse /* do not write unnecessary entries */,
+                                uxrefOffset, outStr, yRef);
+  delete trailerDict;
 
   outStr->close();
   fclose(f);
